class Product
{
  float x, y;
  int number;
  int [] factors;
  int factor_count = 0;
  Bubble [] bubbles;
  
  int bubble_mode;
  float bubble_angle;
  
  float scale = 10.0;
  
  public static final int RANDOM = 0;
  public static final int REGULAR = 1;
  
  Product (int number, float x, float y)
  {
    this.number = number;
    this.x = x;
    this.y = y;
    
    factors = new int[1024];
    
    this.factorise();
    
    bubbles = new Bubble[factor_count];
    
    this.bubble_mode = REGULAR;
    this.bubble_angle = PI / 3;
  }
    
  /*----------------------------------------------------------------------*
   * Naive and shockingly inefficient factorisation routine.
   * Don't care - it's fast enough for the job.
   *----------------------------------------------------------------------*/
  void factorise()
  {
    boolean done = false;
    int product = number;
    
    while (!done)
    {
      int max = (int) sqrt(product);
      int i;
      
      for (i = 2; i <= max; i++)
      {
        if (product % i == 0)
        {
          factors[factor_count++] = i;
          product = product / i;
          break;
        }
      }

      if (i == (max + 1))
      {
        factors[factor_count++] = product;
        done = true;
      }
    }
  }
  
  void _draw()
  {
    for (int i = factor_count - 1; i >= 0; i--)
    {
      float hue = hues[factors[i]];
      fill(hue, 0.8, 1.0);
      int x = i % 20;
      int y = (int) (i / 20);
      ellipse(30 + x * 20, 30 + y * 20, 5 + (sqrt(factors[i])), 5 + (sqrt(factors[i])));
    }
  }
  
  /*----------------------------------------------------------------------*
   * Draw the product as an aggregate of several bubbles,
   * relative to our x and y position.
   *----------------------------------------------------------------------*/  
  void draw()
  {
    float radius = 0f;

    /*----------------------------------------------------------------------*
     * Cycle through each factor bubble, selecting its offset angle (either
     * random, or in increasing quantities of PI / 3) and moving it towards
     * the centre of the cluster until it has collided with another bubble.
     * Inefficient, again, but works nicely and is more foolproof when
     * dealing with random angles.
     *----------------------------------------------------------------------*/     
    for (int i = 0; i < factor_count; i++)
    {
      int j = factor_count - i - 1;
      float angle = 0;

      switch (bubble_mode)
      {
        case RANDOM: angle = random(0, PI * 2); break;
        case REGULAR: angle = i * bubble_angle; break;
      }
      
      radius = scale * (0.5 + (1.5 * log(factors[j])));
      
      boolean collided = false;
      float sx = sin(angle);
      float sy = cos(angle);
      float ax = 0, ay = 0;
      float step = 20 * scale;
      do
      {
        ax = step * sx;
        ay = step * sy;
        
        for (int b = 0; b < i; b++)
        {
          Bubble peer = bubbles[b];
          if (dist(ax, ay, peer.x, peer.y) < (radius + peer.radius))
          {
            collided = true;
            break;
          }
        }
        
        step--;
      } while (!collided && (step >= 0));
      
      Bubble bubble = new Bubble(ax, ay, radius);
      bubbles[i] = bubble;

      /*----------------------------------------------------------------------*
       * Pick hue based on factor's prime index, and draw.
       *----------------------------------------------------------------------*/
      float hue = hues[factors[j]];
      c = hue;
      
      if (DRAW_MONOCHROME)
      {
        stroke(0);
        strokeWeight(0.5);
        noFill();

//        ellipse(x + ax, y + ay, radius * 2, radius * 2);
//        rect(x + ax, y + ay, radius * 2, radius * 2);  
      }
      else
      {
        circle.bind();
        int col = color(hue, 0.9, 0.5);
        renderImage(x + ax, y + ay, radius * 4, radius * 4, red(col), green(col), blue(col));
      }

        // Unused now we are using OpenGL textures (for nice glowing border)
//      ellipse(x + ax, y + ay, radius * 2, radius * 2);
//      rect(x + ax, y + ay, radius * 2, radius * 2);      
    }
    
    /*----------------------------------------------------------------------*
     * If this is a prime number, render a larger flash halo.
     *----------------------------------------------------------------------*/
    if (factor_count == 1)
    {
      noFill();
      stroke(0, 0, 1, 0.4);
      strokeWeight(50);
      ellipse(x, y, 4 + radius * 2, 4 + radius * 2);
      renderImage(x, y, radius * 8, radius * 8, 0.05, 0.05, 0.05);
    }

    /*----------------------------------------------------------------------*
     * Render text on top.
     *----------------------------------------------------------------------*/   
    if (DRAW_TEXT)
    {
      fill(c, 0.4, 1.0, 0.4);
      textFont(font, 3.0 * scale);
      text(number, x, y + (1.0 * scale)); 
    }
  }

  /*----------------------------------------------------------------------*
   * Turn on suitable partials for this product's factors.
   *----------------------------------------------------------------------*/     
  void sound()
  {
    for (int i = 0; i < factor_count; i++)
    {
      int factor = factors[i];
      int factor_index = indices[factor];
      float pan = hues[factor] - 0.5;
      harmonics[factor_index].set("on", 1);
      harmonics[factor_index].set("pan", pan);
    }
  }

  /*----------------------------------------------------------------------*
   * Turn off suitable partials for this product's factors.
   * We do this before the next product is played to avoid having
   * to explicitly switch off *every* partial each time.
   *----------------------------------------------------------------------*/       
  void soundoff()
  {
    for (int i = 0; i < factor_count; i++)
    {
      int factor = factors[i];
      int factor_index = indices[factor];
      harmonics[factor_index].set("on", 0);

    }    
  }
}
